! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: epmips_mod
!
! Authors:  Michael J. Freyberg - MPE Garching - Germany
!          
! Description: Library of EPIC pn specific routines developed at MPE
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module epmips_mod

  use eputils_mod
  use caloaldefs
  use dal
  use oal
  use errorHandling

  implicit none

  private

  public :: SAS_DLIanalysis

contains


! =============================================================================
  subroutine SAS_DLIanalysis(dliarray, evtarray, An_SENDMODE, An_EAMIPSEL, &
                             dli_mean, dli_median, dli_var, dli_sigma, &
                             dli_mean2, dli_median2)

    implicit none
    integer(kind=int32), intent(in), dimension(:) :: dliarray(1:EPN_MAX_Y), &
                                                     evtarray(1:EPN_MAX_Y)
    integer(kind=int32), intent(in)               :: An_EAMIPSEL
    character(len=*),    intent(in)               :: An_SENDMODE
    real(kind=double),   intent(out)              :: dli_mean, dli_median, &
                                                     dli_var, dli_sigma, &
                                                     dli_mean2, dli_median2
    real(kind=double)                             :: evt_mean, evt_median, &
                                                     evt_var, evt_sigma
    integer(kind=int32)                           :: evt_sum, dli_sum
    real(kind=double),   dimension(:), allocatable :: dliarrd, evtarrd
    integer(kind=int32), dimension(:), allocatable :: dlitmp
    integer                                       :: i,nn,mm,kk,iw
    real, parameter                               :: sigma_max = 5.5
    real                                          :: rd, re 
    logical                                       :: dli_warning
    character(len=1024)                           :: dli_warning_text(1:2)
    data dli_warning_text / 'Probably pixel above MIP threshold', &
                            'Probably source with extreme pile-up'/

    dli_warning = .false.
    dli_sum     = 0
    dli_mean    = 0.0d0
    dli_median  = 0.0d0
    dli_mean2   = 0.0d0
    dli_median2 = 0.0d0
    dli_var     = 0.0d0
    dli_sigma   = 0.0d0
    evt_sum     = 0
!
    if (trim(An_SENDMODE) == "MIP CORREC.2") then
      mm = EPN_MAX_X
    elseif (trim(An_SENDMODE) == "MIP CORREC.1") then
      mm = EPN_MAX_Y
!     call message(AppMsg,Verbose,' DLIanalysis: row-wise MIP rejection is unusual.')
    else
      return
    endif
    kk = mm-An_EAMIPSEL
    nn = kk-An_EAMIPSEL

    allocate(dliarrd(1:nn),evtarrd(1:nn),dlitmp(1:mm))
    do i = 1+An_EAMIPSEL,nn+An_EAMIPSEL
      dli_sum = dli_sum + dliarray(i)
      evt_sum = evt_sum + evtarray(i)
    enddo

    call message(AppMsg,Sparse,' ')
    write(errstr,*) 'EPN DLI analysis: MIPSEL = ',An_EAMIPSEL
    call message(AppMsg,Sparse)
    if (verbosity(AppMsg) > Verbose) then
      do i=1,mm
        write(errstr,'(1x,a,i7,i7,i7)') '  DLI for X =',i,dliarray(i),evtarray(i)
        call message(AppMsg,Sparse)
      enddo
    else
      do i=1,An_EAMIPSEL+3
        write(errstr,'(1x,a,i7,i7,i7)') '  DLI for X =',i,dliarray(i),evtarray(i)
        call message(AppMsg,Sparse)
      enddo
      do i=nn+An_EAMIPSEL-2,mm
        write(errstr,'(1x,a,i7,i7,i7)') '  DLI for X =',i,dliarray(i),evtarray(i)
        call message(AppMsg,Sparse)
      enddo
    endif

    dli_mean = dble(dli_sum)/dble(nn)
    dliarrd(1:nn) = dliarray(1+An_EAMIPSEL:nn+An_EAMIPSEL)
    call ep_median0(dliarrd,nn,dli_median)
    dli_var = sqrt(dli_median)
    if (dli_var > 0) then
      dli_sigma = abs(dli_mean-dli_median)/dli_var
    else
      dli_sigma = 0.0
    endif

    dli_mean2 = (dble(dli_sum)-dliarray(1+An_EAMIPSEL)-dliarray(nn+An_EAMIPSEL))/dble(nn-2)
    dliarrd(1:nn-2) = dliarray(2+An_EAMIPSEL:nn+An_EAMIPSEL-1)
!   write(errstr,*) "EPMIPS = ",2+An_EAMIPSEL,nn+An_EAMIPSEL-1,1+(nn+An_EAMIPSEL-1)-(2+An_EAMIPSEL), nn-2
!   call message(AppMsg,Sparse)
    call ep_median0(dliarrd,nn-2,dli_median2)

    evt_mean = dble(evt_sum)/dble(nn)
    evtarrd(1:nn) = evtarray(1+An_EAMIPSEL:nn+An_EAMIPSEL)
    call ep_median0(evtarrd,nn,evt_median)
    evt_var = sqrt(evt_median)
    if (evt_var > 0) then
      evt_sigma = abs(evt_mean-evt_median)/evt_var
    else
      evt_sigma = 0.0
    endif

    write(errstr,'(1x,a,i9,a,f9.1,a,f9.1,a,f9.3,a,f9.1,a,f9.1)') &
                'DLI analysis: sum = ', dli_sum, &
                '  mean = ',dli_mean,'  median = ',dli_median,'  deviation = ',dli_sigma, &
                '  mean2 = ',dli_mean2,'  median2 = ',dli_median2
    call message(AppMsg,Sparse)
    write(errstr,'(1x,a,i9,a,f9.1,a,f9.1,a,f9.3)') 'EVT analysis: sum = ', &
                 evt_sum,'  mean = ',evt_mean,'  median = ',&
                 evt_median,'  deviation = ',evt_sigma
    call message(AppMsg,Sparse)

    dlitmp(1:mm) = dliarray(1:mm)
    if (An_EAMIPSEL == 1) then
      dlitmp(1)  = dlitmp(1)+dli_median*0.20
      dlitmp(mm) = dlitmp(mm)+dli_median*0.20
    elseif (An_EAMIPSEL == 2) then
      dlitmp(1)  = dlitmp(1)+dli_median*0.280
      dlitmp(2)  = dlitmp(2)+dli_median*0.118
      dlitmp(mm-1) = dlitmp(mm-1)+dli_median*0.118
      dlitmp(mm) = dlitmp(mm)+dli_median*0.280
    endif

    if (dli_var > 0) then
!     do i=1+An_EAMIPSEL,nn+An_EAMIPSEL
      do i=1,mm
        rd = abs(dlitmp(i)-dli_median)/dli_var
        if (i.le.An_EAMIPSEL.or.i.gt.kk) then
          rd = rd*0.3
          if (i.lt.An_EAMIPSEL.or.i.gt.kk+1) rd = rd*0.4
        endif
        if (rd > sigma_max) then
          if (.not.dli_warning) call message(AppMsg,Sparse,' ')
          dli_warning = .true.
          re = (evtarray(i)-evt_median)/dli_var
          if (re > sigma_max) then
            iw = 2
          else
            iw = 1
          endif
          write(errstr,*) ' X =',i,'  DLI =',dlitmp(i),'  EVT =',evtarray(i), &
                          '  RD =',nint(rd),'  RE =',nint(re), &
                          '  Reason: ',trim(dli_warning_text(iw))
          call message(AppMsg,Sparse)
        endif
      enddo
    endif
!   if (dli_warning) call warning('DLIanalysis ', &
    if (dli_warning) call message(AppMsg,Sparse,' DLIanalysis: '// &
         'Large inhomogeneity in DLI map. '// &
         'Exposure entries of whole quadrant may be incorrect.')

    call message(AppMsg,Sparse,' ')
    deallocate(dliarrd,evtarrd,dlitmp)

  end subroutine SAS_DLIanalysis



end module epmips_mod
