! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: eputils_mod
!
! Authors:  Michael J. Freyberg - MPE Garching - Germany
!
! Description: Library of EPIC pn specific routines developed at MPE
!
! Comments: 
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!-
module eputils_mod

  use dal

  implicit none

  private 

  public :: ep_median0, nr_moment

  type, public :: EPEAsetupT
    integer            :: lowerthreshold,offset
    integer            :: mipthreshold,mipsel,maxmip
    integer            :: cmcorr      
  end type EPEAsetupT

  type, public :: pnModeInfoT
    integer            :: CalMode
    real(kind=double)  :: frametime, integrationtime
  end type pnModeInfoT

  type, public :: pnEventInfoT
    integer            :: rawX, rawY, amplitude
  end type pnEventInfoT


contains



      subroutine NR_moment(data,n,ave,adev,sdev,var,skew,curt,ierr)
! ----------------------------------------------------------------------
!     Version 1.0 7/8/1990, 1.1 20/8/2012, 2.0 13/9/2017
! ----------------------------------------------------------------------
!     given an inout array of DATA of length N, this routine returns its
!      - arithmetic mean AVE
!      - average deviation ADEV 
!      - standard deviation SDEV
!      - variance VAR
!      - skewness SKEW
!      - kurtosis CURT
!     inspired by NUMERICAL RECIPES
! ----------------------------------------------------------------------

  implicit none

  integer, intent(in)                            :: n
  integer, intent(out)                           :: ierr
  real(kind=double), intent(in), dimension(:)    :: data(1:n)
! real(kind=double), intent(inout), dimension(:) :: x(1:n)
  real(kind=double), intent(out)                 :: ave,adev,sdev,var,skew,curt

! local variables

  integer                                        :: j
  real(kind=double)                              :: sum,power

    ierr = 0
    if (n.le.1) then ! N < 2: too few DATA
      ierr = 1
    else

      sum = 0.d0
      do j = 1,n
        sum = sum + data(j)
      enddo
      
      ave  = sum/n
      adev = 0.d0
      var  = 0.d0
      skew = 0.d0
      curt = 0.d0

      do j = 1,n
        sum   = data(j)-ave
        adev  = adev+abs(sum)
        power = sum*sum
        var   = var+power
        power = power*sum
        skew  = skew+power
        power = power*sum
        curt  = curt+power
      enddo

      adev = adev/n
      var  = var/(n-1)
      sdev = sqrt(var)
      if (var.ne.0.) then
        skew = skew/(n*sdev**3)
        curt = curt/(n*var**2)-3.d0
      else    ! no skew or kurtosis if zero variance
        ierr = 2
      endif

    endif ! sufficient data points, i.e. N>1

    end subroutine NR_moment




  !---------------------------------------------------------------------
  subroutine ep_median0(x, n, xmed)
  !---------------------------------------------------------------------

! Find the median of X(1), ... , X(N), using as much of the quicksort
! algorithm as is needed to isolate it.
! N.B. On exit, the array X is partially ordered.
!
!     Latest revision - 26 November 1996 [alan miller]
!                       [public domain software]
!     minor adaption for XMMSAS 24-05-2002 [mjf]
!
  implicit none

  integer, intent(in)                            :: n
  real(kind=double), intent(inout), dimension(:) :: x(1:n)
  real(kind=double), intent(out)                 :: xmed

! local variables

  real(kind=double)    :: temp, xhi, xlo, xmax, xmin
  logical              :: odd
  integer              :: hi, lo, nby2, nby2p1, mid, i, j, k

  nby2   = n / 2
  nby2p1 = nby2 + 1
  odd    = .true.
  xmed   = 0.0d0

! HI & LO are position limits encompassing the median.

  IF (n == 2 * nby2) odd = .false.
  lo = 1
  hi = n
  IF (n < 3) THEN
    IF (n < 1) THEN
      xmed = 0.0d0
      RETURN
    ENDIF
    xmed = x(1)
    IF (n == 1) RETURN
    xmed = 0.5d0*(xmed + x(2))
    RETURN
  ENDIF

! Find median of 1st, middle & last values.

10 mid = (lo + hi)/2
  xmed = x(mid)
  xlo = x(lo)
  xhi = x(hi)
  IF (xhi < xlo) THEN          ! Swap xhi & xlo
    temp = xhi
    xhi = xlo
    xlo = temp
  ENDIF
  IF (xmed > xhi) THEN
  xmed = xhi
  ELSEIF (xmed < xlo) THEN
    xmed = xlo
  ENDIF

! The basic quicksort algorithm to move all values <= the sort key (XMED)
! to the left-hand end, and all higher values to the other end.

  i = lo
  j = hi
50 DO
    IF (x(i) >= xmed) EXIT
    i = i + 1
  ENDDO
  DO
    IF (x(j) <= xmed) EXIT
    j = j - 1
  ENDDO
  IF (i < j) THEN
    temp = x(i)
    x(i) = x(j)
    x(j) = temp
    i = i + 1
    j = j - 1

! Decide which half the median is in.

    IF (i <= j) GOTO 50
  ENDIF

  IF (.NOT. odd) THEN
    IF (j == nby2 .AND. i == nby2p1) GOTO 130
    IF (j < nby2) lo = i
    IF (i > nby2p1) hi = j
    IF (i /= j) GOTO 100
    IF (i == nby2) lo = nby2
    IF (j == nby2p1) hi = nby2p1
  ELSE
    IF (j < nby2p1) lo = i
    IF (i > nby2p1) hi = j
    IF (i /= j) GOTO 100

! Test whether median has been isolated.

    IF (i == nby2p1) RETURN
  ENDIF
100 IF (lo < hi - 1) GOTO 10
 
  if (.not. odd) then
    xmed = 0.5d0*(x(nby2) + x(nby2p1))
    return
  endif

  temp = x(lo)
  if (temp > x(hi)) then
    x(lo) = x(hi)
    x(hi) = temp
  endif
  xmed = x(nby2p1)
  return

! Special case: N even, J = N/2 & I = J + 1 :
!   so the median is between the two halves of the series.
! Find maximum of the first half and  minimum of the second half, then average.

130 xmax = x(1)
  do k = lo, j
    xmax = MAX(xmax, x(k))
  enddo
  xmin = x(n)

  do k = i, hi
    xmin = MIN(xmin, x(k))
  enddo
  xmed = 0.5d0*(xmin + xmax)

  end subroutine ep_median0




!-----------------------------------------------------------------------
  subroutine PN_framereadout(nevent,Events,EPEAsetup,bad,mipcolumn)
!
! mimic the EPIC-pn frame read-out of one CCD
!
  implicit none
  integer, parameter             :: nx_ccd = 64, ny_ccd = 200
  integer, intent(in)            :: nevent
  type (EPEAsetupT),intent(in)   :: EPEAsetup
  type (pnEventInfoT),intent(in) :: Events(1:nevent)
  logical, intent(out)           :: bad(1:nevent),mipcolumn(1:nx_ccd)
!
  logical       :: start = .true., mipevent, badevent, maskedmode = .false.
  logical       :: badpix(1:nx_ccd,1:ny_ccd) = .false.
  integer       :: iy
  integer       :: ievent, ia, ix, ixx, nmip, npseudo
  integer       :: lower_threshold, mipthreshold, maxmip, mipsel
!
  if (start) then
    start = .false.
    npseudo = 0
    lower_threshold = EPEAsetup%lowerthreshold-EPEAsetup%cmcorr
    mipthreshold    = EPEAsetup%mipthreshold-EPEAsetup%cmcorr
    maxmip          = EPEAsetup%maxmip
    mipsel          = EPEAsetup%mipsel
!   call get_badpixelinfo(badpix)
    if (maskedmode) badpix(33:43,185:195) = .true.
  endif

  mipcolumn(1:nx_ccd)  = .false.
  nmip = 0
  bad(1:nevent) = .false.
!
  do ievent = 1,nevent
    mipevent = .false.
    badevent = .false.
    ix = Events(ievent)%rawX
    iy = Events(ievent)%rawY
    ia = Events(ievent)%amplitude
!
! throw away all events below given amplitude threshold
!
    if (ia .lt. lower_threshold) badevent = .true.
    if (.NOT.badevent) then
!
! here could be a check on bad pixels
!
      if (badpix(ix,iy)) badevent = .true.
!
! now do the check on MIPs
!
      if (.NOT.badevent) then
        if (ia .gt. mipthreshold) then
          mipevent = .true.
          nmip = nmip + 1
          npseudo = npseudo+1
!
! if more than MAXMIP events above MIP threshold: flag all columns
!
          if (nmip .gt. maxmip) then
            mipcolumn(1:nx_ccd) = .true.
            exit
          endif
          if (.not.mipcolumn(ix)) mipcolumn(ix) = .true.
        endif
!
! flag MIPSEL neighbouring columns
!
        if (mipevent) then
          do ixx = ix-1,ix-mipsel,-1
            if (ixx .ge. 1) then
              mipcolumn(ix) = .true.
            else
              exit
            endif
          enddo
          do ixx = ix+1,ix+mipsel,+1
            if (ixx .le. nx_ccd) then
              mipcolumn(ix) = .true.
            else
              exit
            endif
          enddo
        endif
      else  ! badevent badpix
        bad(ievent) = .true.
      endif ! badevent
    else  ! badevent lowthresh
      bad(ievent) = .true.
    endif ! badevent
!
  enddo     ! end of loop over all events in 1 readout frame
!
! flag all events on flagged columns (also neighbouring ones)
!
  do ievent =1,nevent
    ix = Events(ievent)%rawX
    if (mipcolumn(ix)) bad(ievent) = .true.
  enddo
!
! these lines could be moved to main program
!
!     do ix = 1,nx_ccd
!       if (mipcolumn(ix)) nmipcolumn(ix) = nmipcolumn(ix) + 1
!     enddo

!   if (nmip .gt. 0) print*,'Pseudo-MIPs: ',nmip

  end subroutine PN_framereadout


!-----------------------------------------------------------------------
  subroutine get_setup_pn(CalMode,ModeSetup,EPEAsetup)
!
! this routine just initializes a few PN related quantities
! they may be overwritten later with the actual setup (via HK parameters)
!
  implicit none
  integer,            intent(in)  :: CalMode
  type (pnModeInfoT), intent(out) :: ModeSetup
  type (EPEAsetupT),  intent(out) :: EPEAsetup

  ModeSetup%CalMode = CalMode

  print*,'  Determine frametime + integration time ...'
  select case (ModeSetup%CalMode)
    case (0) ! FF
      ModeSetup%frametime = 73.36496D-3
      ModeSetup%integrationtime = 68.70220D-03
    case (1) ! eFF
      ModeSetup%frametime = 199.19408D-3
      ModeSetup%integrationtime = 194.53132D-03
     case (2) ! LW
      ModeSetup%frametime = 47.66384D-3
      ModeSetup%integrationtime = 45.14416D-03
    case (3) ! SW
      ModeSetup%frametime = 5.67180D-3
      ModeSetup%integrationtime = 3.98120D-03
    case (4) ! TI
      ModeSetup%frametime = 5.96464D-3
      ModeSetup%integrationtime = 5.911972D-03
    case (5) ! BU
      ModeSetup%frametime = 4.34448D-3
      ModeSetup%integrationtime = 0.1295958D-03
    case default
      ModeSetup%frametime = 0.0
      ModeSetup%integrationtime = 0.0
  end select

  print*,'  Determine MIP quantities ...'
  if (ModeSetup%CalMode .le. 2) then
    EPEAsetup%mipthreshold = 3512
    EPEAsetup%maxmip = 63
    EPEAsetup%mipsel = 1
    EPEAsetup%lowerthreshold = 532
  elseif (ModeSetup%CalMode .eq. 3) then
    EPEAsetup%mipthreshold = 3512
    EPEAsetup%maxmip = 63
    EPEAsetup%mipsel = 1
    EPEAsetup%lowerthreshold = 532
  elseif (ModeSetup%CalMode .eq. 4) then
    EPEAsetup%mipthreshold = 4096
    EPEAsetup%maxmip = 63
    EPEAsetup%mipsel = 1
    EPEAsetup%lowerthreshold = 552
  elseif (ModeSetup%CalMode .eq. 5) then
    EPEAsetup%mipthreshold = 4096
    EPEAsetup%maxmip = 63
    EPEAsetup%mipsel = 1
    EPEAsetup%lowerthreshold = 535
  else
    EPEAsetup%mipthreshold = 4096
    EPEAsetup%maxmip = 0
    EPEAsetup%mipsel = 0
    EPEAsetup%lowerthreshold = 4096
  endif
  EPEAsetup%cmcorr = 512

  end subroutine get_setup_pn




end module eputils_mod
